#include "config.h"
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <memory.h>

#include "io.h"
#include "io_internal.h"
#include "unix_defs.h"
#include <malloc.h>
#include <string.h>
#include <assert.h>

typedef struct _format_server {
    char *exchange_name;	/* my exchange_name */
    int exch_pid;		/* my exch_pid */
    int port;			/* server port as announced to world */
    char *hostname;		/* server's hostname */
    int conn_sock_inet;		/* public sock for IP socket connection */
    char *usock_name;		/* file name for unix socket connection */
    int portCount;		/* number of current ports */
    FSClient *ports;	/* array of ports index by socket fd */
    fd_set fdset;		/* bitmap of those conns */
    char *data_buffer;		/* buffer for data */
    int buffer_size;
    format_list	*lists[2][1];    /* first dimension byte order, second float format */
} _fsserver;

static FSClient
format_server_accept_conn_sock ARGS((format_server fs, int conn_sock));
static format_server format_server_create();
static void format_server_handle_data ARGS((format_server fs, FSClient fsc));
static int format_server_listen ARGS((format_server fs, int port_num));
static int format_server_poll_and_handle ARGS((format_server fs));

void
main(argc, argv)
int argc; char **argv;
{
    int i;
    format_server fs;
    IOFile test_file;
    char testing_char = 'T';
    int no_fork = 0;

    for (i=1; i<argc; i++) {
	if (strcmp(argv[i], "-no_fork") == 0) {
	    no_fork++;
	} else {
	    fprintf(stderr, "Unknown argument \"%s\"\n", argv[i]);
	    exit(1);
	}
    }
    
    /* test to see if format server is running */
    test_file = new_IOFile();
    establish_server_connection(test_file);
    if (AtomicWrite(test_file->server_fd, &testing_char, 1, test_file) 
	== 1) {
	/* already running */
	exit(0);
    } else {
	if (!no_fork) {
	    if (fork() != 0) {
		/* I'm the parent, return now */
		exit(0);
	    }
	}
    }

    fs = format_server_create();
    if (format_server_listen(fs, FS_PORT) != -1) {
/*	printf("DataExchange server listening at Inet host/port <%s %d>\n",
	       format_server_host_name(fs), format_server_inet_port(fs));*/
	while (1) {
	    format_server_poll_and_handle(fs);
	}
    }
}

static int
format_server_poll_and_handle(format_server fs)
{
    int i, res;
    fd_set rd_set = fs->fdset;

    /* scan existing conn for incoming data & do something */
    res = select(FD_SETSIZE, &rd_set, (fd_set *) NULL,
		 (fd_set *) NULL, NULL);
    if (res == -1) {
	if (errno == EINTR) {
	    return 0;
	}
	fprintf(stderr, "select failed, errno %d\n", errno);
	return -1;
    } else if (res == 0)
	return 0;

    if (fs->conn_sock_inet >= 0 &&
	FD_ISSET(fs->conn_sock_inet, &rd_set)) {
	FD_CLR(fs->conn_sock_inet, &rd_set);
	format_server_accept_conn_sock(fs, fs->conn_sock_inet);
    }
    for (i = 0; i < FD_SETSIZE; i++) {	/* check each possible read fd */
	FSClient fsc = fs->ports[i];
	if (fsc == NULL || !FD_ISSET(i, &rd_set))
	    continue;
	format_server_handle_data(fs, fsc);
	
    }
    return res;
}

/*
typedef enum {
    write_record_state, write_format_state, not_blocked
} block_type;
extern unsigned alarm ARGS((unsigned));
static void record_blocking_state ARGS((block_type state,
					FSClient port));
static int block_count = 0;
static FSClient block_port = NULL;
static block_type block_state = not_blocked;

static void
clear_blocking_state()
{
    block_port = NULL;
    block_state = not_blocked;
    block_count++;
}

static void
record_blocking_state(state, port)
block_type state;
FSClient port;
{
    block_port = port;
    block_state = state;
}

static int last_block_count = 0;

static void
check_blocking_state(signal)
int signal;
{
    if ((block_count == last_block_count) && (block_state != not_blocked)
	&& (block_port != NULL)) {
	fprintf(stderr, "Warning!  No forward progress!!\n");
	fprintf(stderr, " Writing to port [%s, %d] \"%s\".\n",
		block_port->hostname, block_port->port,
		block_port->exchange_name);
    }
    last_block_count = block_count;
    alarm(1);
}


void
DEinit_block_check()
{
    struct sigaction sigact;
    sigact.sa_flags = 0;
    sigact.sa_handler = check_blocking_state;
    sigemptyset(&sigact.sa_mask);
    sigaddset(&sigact.sa_mask, SIGALRM);
    sigaction(SIGALRM, &sigact, NULL);
    alarm(1);
}
*/

static int
format_eq(form1, form2)
IOFormat form1;
IOFormat form2;
{
    int i;
    int all_zero = 1;
    int all_eq = 1;
    for (i=0 ; i< (sizeof(form1->server_ID)/sizeof(form1->server_ID[0])); i++) {
	if (form1->server_ID[i] != 0) {
	    all_zero = 0;
	}
	if (form1->server_ID[i] != form2->server_ID[i]) {
	    all_eq = 0;
	}
    }
    if (all_eq && !all_zero) return 1;
    if (form1->field_count != form2->field_count) return 0;
    if (strcmp(form1->format_name, form2->format_name) != 0) return 0;
    if (form1->record_length != form2->record_length) return 0;
    for (i=0; i< form1->field_count; i++) {
	if (strcmp(form1->field_list[i].field_name,
		   form2->field_list[i].field_name) != 0) return 0;
	if (strcmp(form1->field_list[i].field_type,
		   form2->field_list[i].field_type) != 0) return 0;
	if (form1->field_list[i].field_size != 
	    form2->field_list[i].field_size) return 0;
	if (form1->field_list[i].field_offset != 
	    form2->field_list[i].field_offset) return 0;
    }
    return 1;
}

static IOFormat
find_format(fs, fsc, ioformat)
format_server fs;
FSClient fsc;
IOFormat ioformat;
{
    format_list *list = fs->lists[fsc->byte_reversal][0];
    format_list *last = NULL, *new = NULL;
    while (list != NULL) {
	if (format_eq(ioformat, list->format)) {
	    free_IOformat(ioformat);
	    return list->format;
	}
	last = list;
	list = list->next;
    }
    new = (format_list*)io_malloc(sizeof(format_list));
    new->format = ioformat;
    memcpy(&ioformat->server_ID[0], ioformat, sizeof(ioformat));
    memset(&ioformat->server_ID[0] + sizeof(ioformat), 0, 
	   sizeof(ioformat->server_ID[0]) - sizeof(ioformat));
    new->next = NULL;
    if (last == NULL) {
	fs->lists[fsc->byte_reversal][0] = new;
    } else {
	last->next = new;
    }
    return ioformat;
}

static void
server_write_id(fsc, ioformat)
FSClient fsc;
IOFormat ioformat;
{
    write(fsc->fd, &ioformat->server_ID[0], 8);
}

/* 
 * Close an on-line connection.
 */
static void
FSClient_close(FSClient fsc)
{
    format_server fs = fsc->fs;

    if (fsc->fd >= 0) {
	fs->ports[fsc->fd] = NULL;
	FD_CLR(fsc->fd, &fs->fdset);
	close(fsc->fd);
    }
    io_free(fsc);
}

static void
format_server_handle_data(fs, fsc)
format_server fs;
FSClient fsc;
{
    IOFormat ioformat;
    char	next_action;
    if (AtomicRead(fsc->fd, &next_action, 1, NULL) != 1) {
	FSClient_close(fsc);
	return;
    }
    switch(next_action) {
    case 'T': 
	printf("Testing\n");
	break;
    case 'F':
	printf("Format Registration\n");
	ioformat = server_read_format(fs, fsc);
	dump_IOFormat(ioformat);
	ioformat = find_format(fs, fsc, ioformat);
	server_write_id(fsc, ioformat);
	break;
    case 'G':
	printf("Get Format\n");
	ioformat = io_malloc(sizeof(IOFormatStruct));
	ioformat->field_list = NULL;
	ioformat->var_list = NULL;
	ioformat->field_count = 0;
	ioformat->format_name = NULL;
	ioformat->conversion = NULL;
	server_get_server_ID(fsc->fd, &ioformat->server_ID);
	ioformat = find_format(fs, fsc, ioformat);
	printf("Returning \n");
	dump_IOFormat(ioformat);
	write_format_to_fd(fsc->fd, ioformat);
	break;
    default:
	printf("Uknown request %c\n", next_action);
	break;
    }
    
}

/* 
 * Create and initialize a connection data structure.
 */
static FSClient
FSClient_create(format_server fs)
{
    FSClient fsc = (FSClient) io_malloc(sizeof(*fsc));
    fsc->fd = -1;
    fsc->hostname = NULL;
    fsc->exch_pid = 0;
    fsc->port = 0;
    fsc->fs = fs;
    return fsc;
}

/* 
 * Create and initialize a data exchange.
 */
static format_server
format_server_create()
{
    int len, i, j;
    char name_buf[256];
    format_server fs = (format_server)io_malloc(sizeof(*fs));
    int max_fd = FD_SETSIZE;	/* returns process fd table size */
    if (fs == NULL)
	return NULL;

    /* initialize data structs */
    gethostname(name_buf, 128);
    name_buf[255] = '\0';
    len = strlen(name_buf);
    name_buf[len] = '.';
    getdomainname((&name_buf[len])+1, 128);
    name_buf[255] = '\0';
    len = strlen(name_buf);
    name_buf[len] = '_';
    sprintf(&name_buf[len + 1], "%u", (unsigned int) getpid());
    fs->exchange_name = strdup(name_buf);
    fs->exch_pid = getpid();
    fs->port = -1;
    fs->hostname = NULL;
    fs->conn_sock_inet = -1;
    fs->usock_name = NULL;
    for(i = 0; i<2 ; i++) {
	for (j=0; j<1; j++) {
	    fs->lists[i][j] = NULL;
	}
    }

    fs->portCount = 0;
    fs->ports = (FSClient *) io_malloc(sizeof(FSClient) * max_fd);
    memset((char *) fs->ports, 0, sizeof(FSClient) * max_fd);

    FD_ZERO(&fs->fdset);
    fs->data_buffer = (char*) io_malloc(1);
    fs->buffer_size = 1;
    /* 
     * ignore SIGPIPE's  (these pop up when ports die.  we catch the 
     * failed writes) 
     */
    signal(SIGPIPE, SIG_IGN);

    return fs;
}


/* 
 * Create an IP socket for connection from other data exchanges.
 */
static int
format_server_listen(format_server fs, int port_num)
{
    int length;
    char buf[256];
    char buf2[128];
    struct sockaddr_in sock_addr;
    int sock_opt_val = 1;
    int conn_sock;

    if (fs->conn_sock_inet >= 0) {
	fprintf(stderr, "INET connection socket already open\n");
	return -1;
    }
    conn_sock = socket(AF_INET, SOCK_STREAM, 0);
    if (conn_sock < 0) {
	fprintf(stderr, "Cannot open INET socket\n");
	return -1;
    }
    sock_addr.sin_family = AF_INET;
    sock_addr.sin_addr.s_addr = INADDR_ANY;
    sock_addr.sin_port = htons(port_num);
    if (setsockopt(conn_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &sock_opt_val,
		   sizeof(sock_opt_val)) != 0) {
	fprintf(stderr, "Failed to set 1REUSEADDR on INET socket\n");
	return -1;
    }
    if (bind(conn_sock, (struct sockaddr *) &sock_addr,
	     sizeof sock_addr) < 0) {
	fprintf(stderr, "Cannot bind INET socket\n");
	return -1;
    }
    sock_opt_val = 1;
    if (setsockopt(conn_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &sock_opt_val,
		   sizeof(sock_opt_val)) != 0) {
	perror("Failed to set 2REUSEADDR on INET socket");
	return -1;
    }
    length = sizeof sock_addr;
    if (getsockname(conn_sock, (struct sockaddr *) &sock_addr, &length) < 0) {
	fprintf(stderr, "Cannot get socket name\n");
	return -1;
    }
    fs->port = ntohs(sock_addr.sin_port);

    /* begin listening for conns and set the backlog */
    if (listen(conn_sock, FD_SETSIZE)) {
	fprintf(stderr, "listen failed\n");
	return -1;
    }
    /* record the server's hostname */
    gethostname(buf, 128);
    getdomainname(buf2, 128);
    strcat(buf, ".");
    strcat(buf, buf2);
    fs->hostname = (char *) strdup(buf);

    fs->conn_sock_inet = conn_sock;
    FD_SET(conn_sock, &fs->fdset);
    return 0;
}

/* 
 * Accept socket connection from other data exchs.  If conn_sock is -1,
 * check if there is connection request at conn_sock_inet or
 * conn_sock_unix.
 */
extern FSClient
format_server_accept_conn_sock(format_server fs, int conn_sock)
{
    int sock;
    FSClient fsc;

    struct linger linger_val;

    linger_val.l_onoff = 1;
    linger_val.l_linger = 60;
    if (fs == NULL)
	return NULL;
    if (conn_sock == -1) {
	fd_set fds;
	struct timeval timeout;
	FD_ZERO(&fds);
	if (fs->conn_sock_inet >= 0) {
	    FD_SET(fs->conn_sock_inet, &fds);
	}
	timeout.tv_sec = 0L;
	timeout.tv_usec = 0L;

	if (select(FD_SETSIZE, &fds, (fd_set *) NULL, (fd_set *) NULL,
		   &timeout) > 0) {
	    if (FD_ISSET(fs->conn_sock_inet, &fds)) {
		format_server_accept_conn_sock(fs, fs->conn_sock_inet);
	    }
	}
	return NULL;
    }
    if ((sock = accept(conn_sock, (struct sockaddr *) 0, (int *) 0)) < 0) {
	fprintf(stderr, "Cannot accept socket connection\n");
	return NULL;
    }
    if (setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &linger_val,
		   sizeof(struct linger)) != 0) {
	perror("set SO_LINGER");
	return NULL;
    }

    /* open PBIO descriptor for receiving information */
    fsc = FSClient_create(fs);
    fsc->fd = sock;

    /* open PBIO descriptor for sending information */
    FD_SET(sock, &fs->fdset);
    fs->ports[sock] = fsc;
    fs->portCount++;

    printf("%s: FSxchange_accept_conn_sock successful\n",
	 fs->exchange_name);

    if (fsc->hostname == NULL) fsc->hostname = "none";
    printf("fsc->fd = %d, exch_pid %lx, port = %d, hostname = %s\n", fsc->fd, (long)fsc->exch_pid, fsc->port, fsc->hostname);
    server_read_header(fsc);
    return fsc;
}
