#include "config.h"
#include "assert.h"
#include <stdio.h>
#ifdef STDC_HEADERS
#include <stdlib.h>
#endif
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <sys/types.h>
#include <ctype.h>
#include "io.h"
#include "io_internal.h"
#include "unix_defs.h"

static MAX_INTEGER_TYPE get_big_int ARGS((IOFieldPtr iofield, void *data));
static MAX_FLOAT_TYPE get_big_float ARGS((IOFieldPtr iofield, void *data));
static MAX_UNSIGNED_TYPE get_big_unsigned ARGS((IOFieldPtr iofield, void *data));
static void byte_swap ARGS((char *data, int size));

static int get_double_warn = 0;
static int get_long_warn = 0;

extern void
pbio_internal_convert_field(iofile, src_spec, src, dest_type, dest_size, dest,
	  string_offset_size, string_base, size_delta, converted_strings)
IOFile iofile;
IOFieldPtr src_spec;
void *src;
IOdata_type dest_type;
int dest_size;
void *dest;
int string_offset_size;
char *string_base;
int size_delta;
int converted_strings;
{
    /* quick check to see if it's just a copy... */
    if ((dest_type != string_type) && (dest_type == src_spec->data_type) &&
	(dest_size == src_spec->size)) {
	if (src_spec->byte_swap) {
	    int i;
	    char *destc = (char *) dest;
	    char *srcc = (char *) src + src_spec->offset;
	    for (i = 0; i < dest_size / 2; i++) {
		char tmp = srcc[dest_size - i - 1];
		destc[dest_size - i - 1] = srcc[i];
		destc[i] = tmp;
	    }
	    if (dest_size % 2 != 0) {
		destc[dest_size / 2] = srcc[dest_size / 2];
	    }
	} else {
	    char *srcc = (char *) src + src_spec->offset;
	    memcpy(dest, srcc, dest_size);
	}
	return;
    }
    switch (dest_type) {
    case integer_type:
	{
	    MAX_INTEGER_TYPE tmp = get_big_int(src_spec, src);
	    if (dest_size == sizeof(char)) {
		char *dest_field = (char *) dest;
		*dest_field = tmp;
	    } else if (dest_size == sizeof(short)) {
		short *dest_field = (short *) dest;
		*dest_field = tmp;
	    } else if (dest_size == sizeof(int)) {
		int *dest_field = (int *) dest;
		*dest_field = tmp;
	    } else if (dest_size == sizeof(long)) {
		long *dest_field = (long *) dest;
		*dest_field = tmp;
#if SIZEOF_LONG_LONG != 0
	    } else if (dest_size == sizeof(long long)) {
		long long lltmp = tmp;
		memcpy(dest, &lltmp, sizeof(long long));
#endif
	    } else {
		iofile->result = "size problems in conversion";
	    }
	    break;
	}
    case char_type:
	{
	    char tmp = get_IOchar(src_spec, src);
	    char *dest_field = (char *) dest;
	    *dest_field = tmp;
	    break;
	}
    case enumeration_type:
	{
	    int tmp = get_IOenum(src_spec, src);
	    int *dest_field = (int *) dest;
	    *dest_field = tmp;
	    break;
	}
    case unsigned_type:
	{
	    MAX_UNSIGNED_TYPE tmp = get_big_unsigned(src_spec, src);
	    if (dest_size == sizeof(unsigned char)) {
		unsigned char *dest_field = (unsigned char *) dest;
		*dest_field = tmp;
	    } else if (dest_size == sizeof(unsigned short)) {
		unsigned short *dest_field = (unsigned short *) dest;
		*dest_field = tmp;
	    } else if (dest_size == sizeof(int)) {
		unsigned int *dest_field = (unsigned int *) dest;
		*dest_field = tmp;
	    } else if (dest_size == sizeof(long)) {
		unsigned long *dest_field = (unsigned long *) dest;
		*dest_field = tmp;
#if SIZEOF_LONG_LONG != 0
	    } else if (dest_size == sizeof(long long)) {
		unsigned long long *dest_field = (unsigned long long *) dest;
		*dest_field = tmp;
#endif
	    } else {
		iofile->result = "size problems in conversion";
	    }
	    break;
	}
    case float_type:
	{
	    MAX_FLOAT_TYPE tmp = get_big_float(src_spec, src);
	    if (dest_size == sizeof(float)) {
		float *dest_field = (float *) dest;
		*dest_field = tmp;
	    } else if (dest_size == sizeof(double)) {
		double *dest_field = (double *) dest;
		*dest_field = tmp;
#if SIZEOF_LONG_DOUBLE != 0
	    } else if (dest_size == sizeof(long double)) {
		long double *dest_field = (long double *) dest;
		*dest_field = tmp;
#endif
	    } else {
		iofile->result = "size problems in conversion";
	    }
	    break;
	}
    case string_type:
	if ((dest_size == sizeof(char *)) && (string_base != NULL)) {
	    IOgetFieldStruct tmp_src_field;
	    MAX_INTEGER_TYPE tmp_int;
	    char **dest_field = (char **) dest;
	    tmp_src_field = *src_spec;

	    tmp_src_field.data_type = integer_type;

	    tmp_int = get_big_int(&tmp_src_field, src);


	    if (tmp_int != 0) {
		/* handle possibly different string base */
		tmp_int -= (long) string_offset_size;
		*dest_field = string_base + tmp_int;
	    } else {
		*dest_field = NULL;
	    }
	} else if ((dest_size > sizeof(char *)) && (string_base != NULL)) {
	    /* native field is bigger than char*, store it */
	    IOgetFieldStruct tmp_src_field;
	    MAX_UNSIGNED_TYPE tmp_int;
	    tmp_src_field = *src_spec;

	    tmp_src_field.data_type = integer_type;

	    tmp_int = get_big_unsigned(&tmp_src_field, src);


	    if (tmp_int != 0) {
		/* handle possibly different string base */
		tmp_int -= (long) string_offset_size;
		tmp_int = tmp_int + (MAX_UNSIGNED_TYPE)(unsigned long)string_base;
	    }
	    tmp_src_field.offset = 0;
	    tmp_src_field.size = sizeof(MAX_UNSIGNED_TYPE);
	    tmp_src_field.byte_swap = FALSE;
	    pbio_internal_convert_field(iofile, &tmp_src_field, &tmp_int,
					unsigned_type, dest_size,
					dest, 0, string_base, 0, FALSE);
	} else {
	    /* not a native field struct.  Keep the pointer as an offset. */
	    IOgetFieldStruct tmp_src_field;
	    MAX_INTEGER_TYPE tmp_int;
	    tmp_src_field = *src_spec;

	    tmp_src_field.data_type = integer_type;
	    tmp_int = get_big_int(&tmp_src_field, src);

	    if (!converted_strings) {
		/* record size might change, adjust offset */
		tmp_int += size_delta;
	    }
	    tmp_src_field.offset = 0;
	    tmp_src_field.size = sizeof(MAX_INTEGER_TYPE);
	    tmp_src_field.byte_swap = FALSE;
	    pbio_internal_convert_field(iofile, &tmp_src_field, &tmp_int,
					integer_type, dest_size,
					dest, 0, string_base, 0, FALSE);
	}

	break;
    default:
	assert(FALSE);
    }
}

void
IOconvert_record(conv, src, dest, string_offset_size, final_string_base, converted_strings, src_string_base)
IOConversionPtr conv;
void *src;
void *dest;
int string_offset_size;
void *final_string_base;
int converted_strings;
void *src_string_base;
{
    int i;
    long *control_value = NULL;

    if (src_string_base == NULL) {
	src_string_base = final_string_base;
    }
    for (i = 0; i < conv->conv_count; i++) {
	if (conv->conversions[i].control_field != NULL) {
	    if (control_value == NULL) {
		control_value = (long *) io_malloc(sizeof(long) * conv->conv_count);
	    }
	    control_value[i] = (long)
		get_big_int(conv->conversions[i].control_field, src);
	}
    }
    for (i = 0; i < conv->conv_count; i++) {
	IOFieldPtr src_spec = &conv->conversions[i].src_field;
	int dimen1 = conv->conversions[i].array_dimen1;
	int dimen2 = conv->conversions[i].array_dimen2;

	if (IOhas_error(conv->iofile))
	    return;
	if (!conv->conversions[i].src_field.byte_swap &&
	    (src_spec->size == conv->conversions[i].dest_size) &&
	    (conv->conversions[i].subconversion == NULL) &&
	    (conv->conversions[i].control_field == NULL) &&
	    ((src_spec->data_type != string_type) || (final_string_base == NULL))) {
	    /* data movement is all that is required */
	    void *dest_field = (void *) (conv->conversions[i].dest_offset +
					 (char *) dest);
	    memcpy(dest_field, (char *) src + src_spec->offset,
		   src_spec->size * dimen1 * dimen2);
	} else {
	    int j;
	    void *dest_field = (void *) (conv->conversions[i].dest_offset +
					 (char *) dest);
	    IOgetFieldStruct tmp_spec;
	    IOdata_type dest_type = src_spec->data_type;
	    int dest_size = conv->conversions[i].dest_size;
	    tmp_spec = *src_spec;

	    if (conv->conversions[i].control_field != NULL) {
		/* really a variant array, adjust address like a string */
		tmp_spec.data_type = string_type;
		tmp_spec.size = conv->ioformat->pointer_size;
		dest_type = string_type;
		dest_size = conv->target_pointer_size;
		if (conv->conversion_type == copy_strings) {
		    IOgetFieldStruct tmp_spec;
		    char *final_addr;
		    long tmp;
		    tmp_spec = *src_spec;
		    tmp_spec.size = conv->ioformat->pointer_size;
		    tmp_spec.data_type = integer_type;
		    tmp_spec.byte_swap = conv->ioformat->byte_reversal;
		    tmp = get_IOlong(&tmp_spec, src);
		    final_addr = (char *) final_string_base + tmp
			- string_offset_size;
		    if (((unsigned long) final_addr % 8) != 0) {
			int align_size = 8 - ((unsigned long) final_addr % 8);
			/* printf("Adjust by %d\n", align_size); */
			final_string_base = (char *) final_string_base +
			    align_size;
		    }
		}
	    }
	    for (j = 0; j < dimen1 * dimen2; j++) {
		tmp_spec.offset = src_spec->offset + j * tmp_spec.size;
		if ((conv->conversions[i].control_field != NULL) ||
		    (conv->conversions[i].subconversion == NULL)) {
		    /* simple (native) field or variant array */
		    pbio_internal_convert_field(conv->iofile, &tmp_spec, src,
						dest_type, dest_size,
					  dest_field, string_offset_size,
						final_string_base,
						conv->base_size_delta,
						converted_strings);
		} else {
		    IOConversionPtr subconv =
		    conv->conversions[i].subconversion;
		    void *subsrc = (void *) ((char *) src + tmp_spec.offset);
		    IOconvert_record(subconv, subsrc, dest_field,
				   string_offset_size, final_string_base,
				     converted_strings, src_string_base);
		}
		dest_field = (void *) ((char *) dest_field +
				       conv->conversions[i].dest_size);
	    }
	}
	if (((conv->conversion_type == copy_strings) &&
	     (conv->conversions[i].src_field.data_type == string_type)) ||
	    (conv->conversions[i].control_field != NULL)) {

	    char *final_area = NULL, *src_area = NULL;
	    int copy_length = 0;

	    /* variant array or string to be copied case */
	    if (final_string_base != NULL) {
		IOgetFieldStruct tmp_spec;
		long tmp;
		char **dest_field;
		tmp_spec = *src_spec;
		tmp_spec.size = conv->ioformat->pointer_size;
		tmp_spec.data_type = integer_type;
		if (conv->conversion_type != direct_to_mem) {
		    /* conversion must be done */
		    tmp_spec.byte_swap = conv->ioformat->byte_reversal;
		} else {
		    /* conversion has already taken place */
		    tmp_spec.byte_swap = 0;
		}
		tmp = get_IOlong(&tmp_spec, src);

		if ((tmp != 0) && (conv->conversion_type != direct_to_mem)) {
		    /* handle possibly different string base */
		    tmp -= (long) string_offset_size;
		    src_area = (char *) src_string_base + tmp;
		    final_area = (char *) final_string_base + tmp;
		} else {
		    src_area = (char *) tmp;
		    final_area = (char *) tmp;
		}
		dest_field = (char **) (conv->conversions[i].dest_offset +
					(char *) dest);
	    } else {
		/* not a native field struct.  Keep the pointer as an *
		 * offset. */
		IOgetFieldStruct tmp_src_field;
		MAX_INTEGER_TYPE tmp_int;
		void *dest_field = (void *) (conv->conversions[i].dest_offset +
					     (char *) dest);

		tmp_src_field = *src_spec;
		tmp_src_field.data_type = integer_type;
		tmp_src_field.size = conv->ioformat->pointer_size;
		tmp_int = get_big_int(&tmp_src_field, src);

		if (!converted_strings) {
		    /* record size might change, adjust offset */
		    tmp_int += conv->base_size_delta;
		}
		tmp_src_field.offset = 0;
		tmp_src_field.size = sizeof(MAX_INTEGER_TYPE);
		tmp_src_field.byte_swap = FALSE;
		pbio_internal_convert_field(conv->iofile, &tmp_src_field,
					    &tmp_int, integer_type,
					  conv->conversions[i].dest_size,
					  dest_field, string_offset_size,
					    final_string_base, 0, FALSE);
	    }
	    /* handle copying or conversion */
	    if (conv->conversions[i].src_field.data_type == string_type) {
		if (src_area != NULL)
		    copy_length = strlen(src_area) + 1;
	    } else {
		/* variant array */
		if (!conv->conversions[i].src_field.byte_swap &&
		    (src_spec->size == conv->conversions[i].dest_size) &&
		    (conv->conversions[i].subconversion == NULL) &&
		    ((src_spec->data_type != string_type) ||
		     (final_string_base == NULL))) {

		    /* data movement is all that is required */
		    if (conv->conversion_type == copy_strings) {
			copy_length = conv->conversions[i].dest_size *
			    control_value[i];
		    } else {
			/* not even that is necessary */
		    }
		} else {
		    /* must do conversions */
		    int length = control_value[i];
		    int j;
		    void *dest_field = final_area;
		    IOgetFieldStruct tmp_spec;
		    IOdata_type dest_type = src_spec->data_type;
		    int dest_size = conv->conversions[i].dest_size;

		    tmp_spec = *src_spec;
		    final_string_base = (char *) final_string_base +
			(length * (dest_size - src_spec->size));
		    for (j = 0; j < length; j++) {
			tmp_spec.offset = j * src_spec->size;
			if (conv->conversions[i].subconversion == NULL) {
			    /* simple (native) field in variant array */
			    pbio_internal_convert_field(conv->iofile,
						     &tmp_spec, src_area,
						    dest_type, dest_size,
							dest_field,
						      string_offset_size,
							final_string_base,
						   conv->base_size_delta,
						      converted_strings);
			} else {
			    IOConversionPtr subconv =
			    conv->conversions[i].subconversion;
			    void *subsrc = src_area + j * src_spec->size;
			    IOconvert_record(subconv, subsrc, dest_field,
					     string_offset_size,
					     final_string_base,
					     converted_strings,
					     src_string_base);
			}
			dest_field = (void *) ((char *) dest_field +
					 conv->conversions[i].dest_size);
		    }
		}
	    }
	    if (copy_length != 0) {
		memcpy(final_area, src_area, copy_length);
	    }
	}
    }
    if (control_value != NULL) {
	io_free(control_value);
    }
}

static MAX_INTEGER_TYPE
get_big_int(iofield, data)
IOFieldPtr iofield;
void *data;
{
    if (iofield->data_type == integer_type) {
	if (iofield->size == sizeof(char)) {
	    char tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(char));
	    return (long) tmp;
	} else if (iofield->size == sizeof(short)) {
	    short tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(short));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(short));
	    return (long) tmp;
	} else if (iofield->size == sizeof(int)) {
	    int tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(int));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(int));
	    return (long) tmp;
	} else if (iofield->size == sizeof(long)) {
	    long tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(long));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(long));
	    return tmp;
	} else if (iofield->size == 2 * sizeof(long)) {
	    long tmp;
	    int low_bytes_offset = iofield->offset;
#ifdef WORDS_BIGENDIAN
	    if (!iofield->byte_swap) {
		low_bytes_offset += sizeof(long);
	    }
#else
	    if (iofield->byte_swap) {
		low_bytes_offset += sizeof(long);
	    }
#endif
	    memcpy(&tmp, (char *) data + low_bytes_offset, sizeof(long));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(long));
	    return tmp;
	} else {
	    if (!IO_shut_up && !get_long_warn) {
		fprintf(stderr, "Get Long failed!  Size problems.  File int size is %d.\n", iofield->size);
		get_long_warn++;
	    }
	    return -1;
	}
    } else if (iofield->data_type == unsigned_type) {
	MAX_UNSIGNED_TYPE tmp = get_big_unsigned(iofield, data);
	return (MAX_UNSIGNED_TYPE) tmp;
    } else if (iofield->data_type == float_type) {
	MAX_FLOAT_TYPE tmp = get_big_float(iofield, data);
#ifndef METICULOUS_FLOATS_AND_LONGS
	return (MAX_INTEGER_TYPE) (long) (double) tmp;
#else
	return (MAX_INTEGER_TYPE) tmp;
#endif
    } else {
	fprintf(stderr, "Get IOlong failed on invalid data type!\n");
	exit(1);
    }
    /* NOTREACHED */
    return 0;
}

static MAX_UNSIGNED_TYPE
get_big_unsigned(iofield, data)
IOFieldPtr iofield;
void *data;
{
    if (iofield->data_type == unsigned_type) {
	if (iofield->size == sizeof(char)) {
	    unsigned char tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(char));
	    return (MAX_UNSIGNED_TYPE) tmp;
	} else if (iofield->size == sizeof(short)) {
	    unsigned short tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(short));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(short));
	    return (MAX_UNSIGNED_TYPE) tmp;
	} else if (iofield->size == sizeof(int)) {
	    unsigned int tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(int));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(int));
	    return (MAX_UNSIGNED_TYPE) tmp;
	} else if (iofield->size == sizeof(long)) {
	    unsigned long tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(long));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(long));
	    return tmp;
	} else if (iofield->size == 2 * sizeof(long)) {
	    unsigned long tmp;
	    int low_bytes_offset = iofield->offset;
#ifdef WORDS_BIGENDIAN
	    if (!iofield->byte_swap) {
		low_bytes_offset += sizeof(long);
	    }
#else
	    if (iofield->byte_swap) {
		low_bytes_offset += sizeof(long);
	    }
#endif
	    memcpy(&tmp, (char *) data + low_bytes_offset, sizeof(long));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(long));
	    return tmp;
	} else {
	    if (!IO_shut_up && !get_long_warn) {
		fprintf(stderr, "Get Long failed!  Size problems.  File int size is %d.\n", iofield->size);
		get_long_warn++;
	    }
	    return -1;
	}
    } else if (iofield->data_type == integer_type) {
	MAX_INTEGER_TYPE tmp = get_big_int(iofield, data);
	return (MAX_UNSIGNED_TYPE) tmp;
    } else if (iofield->data_type == float_type) {
	MAX_FLOAT_TYPE tmp = get_big_float(iofield, data);
#ifndef METICULOUS_FLOATS_AND_LONGS
	return (MAX_UNSIGNED_TYPE) (long) (double) tmp;
#else
	return (MAX_UNSIGNED_TYPE) tmp;
#endif
    } else {
	fprintf(stderr, "Get IOulong failed on invalid data type!\n");
	exit(1);
    }
    /* NOTREACHED */
    return 0;
}

static MAX_FLOAT_TYPE
get_big_float(iofield, data)
IOFieldPtr iofield;
void *data;
{
    if (iofield->data_type == float_type) {
	if (iofield->size == sizeof(float)) {
	    float tmp;
	    MAX_FLOAT_TYPE tmp2;
	    memcpy(&tmp, ((char *) data + iofield->offset), sizeof(float));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(float));
	    tmp2 = tmp;
	    return tmp2;
	} else if (iofield->size == sizeof(double)) {
	    double tmp;
	    MAX_FLOAT_TYPE tmp2;
	    memcpy(&tmp, ((char *) data + iofield->offset), sizeof(double));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(double));
	    tmp2 = tmp;
	    return tmp2;
#if SIZEOF_LONG_DOUBLE != 0
	} else if (iofield->size == sizeof(long double)) {
	    long double tmp;
	    MAX_FLOAT_TYPE tmp2;
	    memcpy(&tmp, ((char *) data + iofield->offset),
		   sizeof(long double));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp,
			  sizeof(long double));
	    tmp2 = tmp;
	    return tmp2;
#endif
	} else {
	    if (!IO_shut_up && !get_double_warn) {
		fprintf(stderr, "Get Double failed!  Size problems.  File double size is %d.\n", iofield->size);
		get_double_warn++;
	    }
	    return 0.0;
	}
    } else if (iofield->data_type == integer_type) {
	MAX_INTEGER_TYPE tmp = get_big_int(iofield, data);
#ifndef METICULOUS_FLOATS_AND_LONGS
	/* 
	 * A concession to inter-compiler interoperability...  We
	 * shouldn't need the (double)(long) casts here.  If we don't
	 * use them, AND we're on a machine which doesn't support "long 
	 * long" and "long double" in native code, gcc generates calls 
	 * to libgcc and the resulting library can't be linked with
	 * anything but GCC.  This can be a problem for installed
	 * libraries.  Using the casts avoids that particular problem,
	 * at a cost of data loss in the case of someone converting a
	 * "long long" that doesn't fit in a long to a floating point
	 * value. 
	 */
	return (MAX_FLOAT_TYPE) (double) (long) tmp;
#else
	return (MAX_FLOAT_TYPE) tmp;
#endif
    } else if (iofield->data_type == unsigned_type) {
	MAX_UNSIGNED_TYPE tmp = get_big_unsigned(iofield, data);
#ifndef METICULOUS_FLOATS_AND_LONGS
	return (MAX_FLOAT_TYPE) (double) (long) tmp;
#else
	return (MAX_FLOAT_TYPE) tmp;
#endif
    } else {
	fprintf(stderr, "Get Double failed on invalid data type!\n");
	exit(1);
    }
    /* NOTREACHED */
    return 0;
}

static void
byte_swap(data, size)
char *data;
int size;
{
    int i;
    assert((size % 2) == 0);
    for (i = 0; i < size / 2; i++) {
	char tmp = data[i];
	data[i] = data[size - i - 1];
	data[size - i - 1] = tmp;
    }
}
