
#define MAGIC_NUMBER 0x4356ffa9	/* random magic */
#define REVERSE_MAGIC_NUMBER 0xa9ff5643		/* byte reversed random
						 * magic */
#define MAGIC_FLOAT 0.0078125	/* random float */
#define CURRENT_IO_VERSION 5
#ifndef FALSE
#define FALSE 0
#define TRUE 1
#endif

#define FS_PORT 5347

extern char *architecture;

typedef enum {
    unknown_type, integer_type, unsigned_type, float_type,
    char_type, string_type, enumeration_type
} IOdata_type;

typedef enum {
    Format_Unknown= 10, Format_IEEE= 0  /* more later? */
} IOfloat_formats;

typedef struct _IOFileStruct {
    int file_id;
    int IOversion;
    IOStatus status;
    char *architecture;
    int reg_format_count;
    int byte_reversal;
    int pointer_size;
    int errno;
    char *result;

/* used for contexts but not true IOFile's */
    int is_context;
    int server_fd;
    int server_byte_reversal;
    IOConversionCallbackFunc conversion_callback;

    int read_ahead;
    IORecordType next_record_type;
    IOFormat next_record_format;
    int next_record_len;
    int remaining_count;
    void *tmp_buffer;
    int tmp_buffer_size;
    int tmp_buffer_in_use_size;
    int format_list_size;
    IOFormat *format_list;
} IOFileStruct;

typedef struct _IOVarInfoStruct {
    int string;
    int var_array;
    IOFieldPtr control_field;
    IOFormat subformat;
} IOVarInfoStruct, *IOVarInfoList;

typedef struct _IOFormatStruct {
    char *format_name;
    int format_id;
#if SIZEOF_INT == 2
    int server_ID[4];
#endif
#if SIZEOF_INT == 4
    int server_ID[2];
#endif
#if SIZEOF_INT == 8
    int server_ID[1];
#endif
    int record_length;
    int byte_reversal;
    int pointer_size;
    int IOversion;
    IOFile iofile;
    IOConversionPtr conversion;
    int field_count;
    int variant;
    IOFieldList field_list;
    IOVarInfoList var_list;
} IOFormatStruct;

typedef struct _IOgetFieldStruct {
    int offset;
    int size;
    IOdata_type data_type;
    int byte_swap;
} IOgetFieldStruct;

typedef enum {
    none_required, direct_to_mem, buffer_and_convert, copy_strings
} IOconversion_type;

typedef struct _IOconvFieldStruct {
    IOgetFieldStruct src_field;
    int array_dimen1, array_dimen2;
    IOFieldPtr control_field;
    int dest_offset;
    int dest_size;
    IOConversionPtr subconversion;
} IOconvFieldStruct;

typedef struct _IOConversionStruct {
    IOconversion_type conversion_type;
    int notify_of_format_change;
    IOFieldList native_field_list;
    int conv_count;
    int base_size_delta;	/* native size - file record length */
    double max_var_expansion;
    int target_pointer_size;
    IOFile iofile;
    IOFormat ioformat;
    IOconvFieldStruct conversions[1];
} IOConversionStruct;

extern int IO_shut_up;

#if SIZEOF_INT == 4
#define FILE_INT int
#else
Whoops...  Unhandled.
#endif

extern void *io_malloc ARGS((int size));
extern void io_free ARGS((void *ptr));
extern void *io_realloc ARGS((void *ptr, int size));
extern IOdata_type str_to_data_type ARGS((char *str));
extern IOdata_type array_str_to_data_type ARGS((char *str, int *dimen1,
						int *dimen2));
extern char *data_type_to_str ARGS((IOdata_type data_type));
extern int field_type_eq ARGS((char *str1, char *str2));
extern int field_offset_compar ARGS((const void *a, const void *b));
extern int AtomicRead ARGS((int fd, void *buffer, int length, IOFile iofile));
extern int AtomicWrite ARGS((int fd, void *buffer, int length, IOFile iofile));
extern int get_AtomicInt ARGS((IOFile iofile, FILE_INT * buffer));
extern int put_AtomicInt ARGS((IOFile iofile, FILE_INT * buffer));

extern void IOconvert_record ARGS((IOConversionPtr conv, void *src,
				   void *dest, int string_offset_size,
				   void *final_string_area,
				   int converted_strings, 
				   void* src_string_area));
extern void reset_read_ahead ARGS((IOFile iofile));
extern void expand_IOFile ARGS((IOFile iofile));
extern int write_IOformat ARGS((IOFile iofile, IOFormat ioformat));
extern int sdump_value ARGS((char *str, char *field_type, int field_size,
			     int field_offset, IOFile iofile, void *data,
			     void *string_base, int byte_reversal));
extern int count_IOfield ARGS((IOFieldList field_list));
extern char *get_IOstring_base ARGS((IOFieldPtr iofield, void *data, void *string_base));
extern char *base_data_type ARGS((char *data_type));
extern void dump_IOConversion ARGS((IOConversionPtr conv_ptr));
extern int get_var_array_control ARGS((char *str, IOFieldList fields));
extern int min_align_size ARGS((int size));
extern void generate_var_list ARGS((IOFormat ioformat));
extern void pbio_internal_convert_field ARGS((IOFile iofile, 
					      IOFieldPtr src_spec, void *src,
					      IOdata_type dest_type, 
					      int dest_size, void *dest, 
					      int string_offset_size,
					      char *string_base, 
					      int size_delta,
					      int converted_strings));

extern void 
set_IOconversion_for_format ARGS((IOFile iofile, IOFormat file_ioformat,
				  IOFieldList native_field_list,
				  int native_struct_size));

#define Max(i,j) ((i<j) ? j : i)

/* negative IOformat numbers are reserved as escape codes! */
#define ARRAY_FOLLOWS -1
#define COMMENT_FOLLOWS	-2
#define FORMAT_FOLLOWS -3
#define REREGISTERED_FORMAT_FOLLOWS -4

#if SIZEOF_LONG_DOUBLE != 0 && SIZEOF_LONG_DOUBLE != SIZEOF_DOUBLE
#define MAX_FLOAT_TYPE long double
#else
#define MAX_FLOAT_TYPE double
#define MAX_FLOAT_GET get_IOdouble
#endif
#if SIZEOF_LONG_LONG != 0
#define MAX_INTEGER_TYPE long long
#else
#define MAX_INTEGER_TYPE long
#endif

#define MAX_UNSIGNED_TYPE unsigned MAX_INTEGER_TYPE

struct _format_server;

typedef struct _format_server *format_server;

typedef struct _FSclient {
    char *exchange_name;		/* port's exchange_name */
    int exch_pid;			/* port's exch_pid */
    int port;			/* port's port number */
    char *hostname;		/* port's host name */
    char *usock_name;		/* port's unix socket name */
    int fd;
    int byte_reversal;
    format_server fs;
} *FSClient;

typedef struct format_list {
    IOFormat	format;
    time_t	last_reference;
    struct format_list *next;
} format_list;

extern IOFile new_IOFile ARGS((void));
extern IOFormat server_read_format();
extern void server_write_header ARGS((IOFile iofile));
extern int write_format_to_fd ARGS((int fd, IOFormat ioformat));
extern void free_IOformat ARGS((IOFormat format));
extern void server_get_server_ID ARGS((int fd, void *server_ID));
extern void server_read_header ARGS((FSClient fsc));
extern void establish_server_connection ARGS((IOFile iofile));
